//              CS 290-Java Demo 4
//   The Eight Non-Attacking Queens Problem
//
//           By Charles W. Neville
//    Copyright 1998 by Charles W. Neville
//
//   Based on many similar solutions, in, for example, Data Structures and 
//   Program Design in Pascal, 2nd ed., by Larry Nyhoff and Sanford Leestma, 
//   MacMillan, 1992, or in An Introduction to Object Oriented Programming, 
//   by Timothy Budd, Addison Wesley, 1991.

// The classical chess problem of the Eight Non-Attacking Queens asks for a 
// configuration of eight queens on an 8 x 8 chessboard so that no queen can 
// attack any other. The algorithm for solving this problem proceeds 
// recursively:  First solve the subproblem of finding a non-attacking 
// configuration for the 7 queens in columns 0 -- 7.  Then move the remaining 
// queen up row by row, starting at row 0, until she cannot too attack any of 
// her neighbors.  If she is moved all the way to row 7 and still can attack, 
// move her neighbor to the immediate left up one row and start the whole 
// solution process over.
//
// Although the basic algorithm is the same in all languages permitting 
// recursion, the form of the solution in an object oriented language such as 
// Java differs in that the queens themselves can be implemented as objects, 
// and each queen decides where to move next.  The queens thus act as though 
// they were endowed with a rudimentary intelligence, and they cooperate with 
// one another to solve the problem.  By contrast, the form of the solution in 
// a non-object oriented language treats the queens as inanimate objects which 
// are moved by separate block of code implementing the algorithm.
//
// One can argue that the form of the non-object oriented solution more closely 
// mimics the way a human chess player would proceed, because the person would 
// make all decisions himself or herself, and then act on the chess pieces by 
// moving them around. The object oriented solution, on the other hand, is very 
// interesting because it uses the twin principals that objects have behavior 
// and can communicate, and it uses this pair of principals to create an 
// interacting community of objects which mutually arrive at a solution.
//
// This program only finds one solution to the Eight Non-Attacking Queens 
// problem.  There are many solutions, in fact a huge number, and a somewhat 
// more difficult program enumerates all of them.  Finding one solution by 
// hand, without the aid of a computer is easy, but it would seem to be an 
// impossible task to find ALL solutions by hand.  So it is quite remarkable 
// that the great mathematician Carl Friedrich Gauss (1777-1855) enumerated all 
// solutions by hand in the late 18th century!  You can read a translation of 
// his original paper in "The World of Mathematics", James R. Newman editor.  
// You might also want to read Newman's commentary on Gauss, titled "The Prince 
// of Mathematicians", in the same volume.
//
// To return to this program, it is also interesting in that it employs a 
// linked list of eight queen objects.  Each queen knows who her left hand 
// neighbor is via an object reference variable (an implicit pointer) named 
// leftQueen.  (The leftQueen pointer takes the place of the next pointer in a 
// classical linked list.)  Each queen can ask her left neighbor to advance via 
// a method call such as leftQueen.advanceQueen(), and she can ask her left 
// neighbor to solve the left neighbors' non-attacking queens sub-problem via a 
// method call such as leftQueen.solveNonAttackingQueens().

import java.awt.*;
import java.applet.Applet;

// A chess queen with the minimal behavior needed
// to solve the Eight Queens problem.
class Queen {

    // position of queen
    int row;  // row number 0 to 7
    int col;  // column number 0 to 7

    Queen leftQueen;  // we are going to have a linked list
                      // of queens going from right to left

    // constructor
    Queen(int row, int col) {

         leftQueen = null;
         this.row = row;
         this.col = col;
    }

    // get left neighbor
    Queen getLeftQueen() {

       return leftQueen;
    }

    // set left neighbor
    void setLeftQueen(Queen q) {

       leftQueen = q;
    }

    // get the row
    int getRow() {

       return row;
    }

    // get the column
    int getCol() {

       return col;
    }

    // advance queen by 1 row
    void advanceQueen() {

       if (row < 7) row++;
       else row = 0;
    }

    // can this queen be attacked by a designated queen q to her right?
    boolean canBeAttacked(Queen q) {

        int r = q.getRow();
        int c = q.getCol();

        // same row?
        if (r == row) return  true;

        // on diagonal?
        if (r - row == c - col) return true;
        if (r - row == col - c) return true;

        // because each queen advances in ONE column, there will never be 
        // another queen in the same column.  Thus, if we get to here, this 
        // queen cannot be attacked by queen q
        return false;
     }

     // can this queen attack any queen to her left?
     boolean canAttack() {

           Queen nextQueen = leftQueen;

           while( (nextQueen != null) && !nextQueen.canBeAttacked(this) )
               nextQueen = nextQueen.getLeftQueen();

           if (nextQueen == null) return false;

           return true;
     }

     // recursively solve the non-attacking queens problem
     // for this queen and the queens to her left
     void solveNonAttackingQueens() {

        if ( leftQueen == null) return;  // recursive exit first!

        // if this queen has advanced all the way to row 7 and still can
        // attack, advance the leftQueen to alter the left configuration
        // before re-solving the left non-attacking queens subproblem
        if (getRow() == 7) leftQueen.advanceQueen();
        leftQueen.solveNonAttackingQueens();

        // if this queen can still attack, advance her by 1 row and try again
        if ( canAttack() ) {
           advanceQueen(); 
           solveNonAttackingQueens();
        }
      }

     // draw this queen starting at her upper left corner
     void drawQueen(Graphics g) {

        // starting point in screen X, Y coordinates
        int startX = col*20 + 2;
        int startY = row*20 + 2;

        // build a polygon in the shape of an 18 pixel by 18 pixel crown
        Polygon crown = new Polygon();
        crown.addPoint(startX, startY);
        crown.addPoint(startX + 9, startY + 9);
        crown.addPoint(startX + 18, startY);
        crown.addPoint(startX + 18, startY + 18);
        crown.addPoint(startX, startY + 18);
        // polygons are automatically closed, so don't need to
        // add starting point again

        // and draw it
        g.drawPolygon(crown);
        g.fillPolygon(crown);
     } 
}


// A chess board which contains eight queens,
// and can draw itself and the eight queens
class QueensBoard {

     Queen rightQueen;  // rightmost queen in linked list of eight queens

     // constructor
     QueensBoard() {

        Queen currentQueen, nextQueen;

        rightQueen = new Queen(0, 7);  // rightmost queen in row 0 col 7
        currentQueen= rightQueen;

        // rest of queens all in row 0
        for(int col = 6; col >= 0; col--) {

            nextQueen = new Queen(0, col);
            currentQueen.setLeftQueen(nextQueen);
            currentQueen = nextQueen;
        }

        // Note: now have linked list of queens, with leftmost
        // queen's leftQueen pointer set at null
     }

     // get the rightmost queen
     Queen getRightQueen() {

         return rightQueen;
     }

      // recursively draw the eight queens
      void drawQueens(Queen q, Graphics g) {

           if (q != null) {

              drawQueens(q.getLeftQueen(), g);
              q.drawQueen(g);
           }
      }

      void drawBoard(Graphics g) {

          // draw column lines
          for (int col = 0; col <= 8; col++) {

              g.drawLine(0, 20*col, 160, 20*col);
          }

          // draw row lines
          for (int row = 0; row <= 8; row++) {

              g.drawLine(20*row, 0, 20*row, 160);
          }

          // draw the queens
          drawQueens(rightQueen, g);
      }

      void clearBoard(Graphics g) {

         g.setColor(Color.white);
         g.fillRect(0, 0, 160, 160);
         g.setColor(Color.black);
      }
}

public class EightQueens extends Applet {

   Queen rightQueen;
   QueensBoard board;

   public void init() {

       board = new QueensBoard();
       rightQueen = board.getRightQueen();
       setSize(160, 180);
   }

   public void start() {

       rightQueen.solveNonAttackingQueens();
   }

   public void paint(Graphics g) {

       board.clearBoard(g);
       board.drawBoard(g);
   }
}